/*
 * Created on 28. 7. 2017 9:44:08
 *
 */
package com.jh.common.security.controller;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;
import javax.servlet.http.HttpSession;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.condition.ConditionalOnExpression;
import org.springframework.context.ApplicationListener;
import org.springframework.context.MessageSource;
import org.springframework.context.event.ContextRefreshedEvent;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Controller;
import org.springframework.util.StringUtils;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseStatus;

import com.jh.common.security.AuthService;
import com.jh.common.security.PasswordSecurity;
import com.jh.common.security.ResetPassword;
import com.jh.common.security.Signup;
import com.jh.common.template.TemplateService;
import com.jh.common.web.error.NotFoundException;

/**
 * The Class AuthController.
 *
 * @author Jan Hrabal
 */
@Controller
@ConditionalOnExpression("${auth.controller.enabled:true}")
public class AuthController implements ApplicationListener<ContextRefreshedEvent> {
    
	/** The style href. */
	@Value("${auth.html.style.href:/p.css}")
	private String styleHref;
	
    /** The content encoding. */
    @Value("${public.content.encoding:UTF-8}")
    private String contentEncoding;	
	
    /** The template service. */
    @Autowired
    private TemplateService templateService;
    
    /** The auth service. */
    @Autowired
    private AuthService authService;
    
    /** The message source. */
    private MessageSource messageSource;
    
    @Value("${auth.controller.signup.enabled:true}")
    private boolean signupEnabled;
    
    @Value("${auth.controller.reset.enabled:true}")
    private boolean resetEnabled;
    
    @Value("${auth.controller.home.enabled:true}")
    private boolean homeEnabled;    
    
    private static boolean signupEnabledStatic = true;
    private static boolean resetEnabledStatic = true;
    private static boolean homeEnabledStatic = true;
    
    
    /**
     * Inits the.
     */
    @PostConstruct
    public void init() {
        if (contentEncoding != null) {
            contentEncoding = contentEncoding.toLowerCase();
        } else {
            contentEncoding = "utf-8";
        }
    }
    
    
    /**
     * Logout.
     *
     * @param session the session
     * @return the string
     */
    @RequestMapping({"logout", "/auth/logout"})
    public String logout(HttpSession session) {
        doLogout(session);
        return "redirect:/";
    }


	/**
	 * Do logout.
	 *
	 * @param session the session
	 */
	protected void doLogout(HttpSession session) {
		SecurityContextHolder.clearContext();
        if (session != null) {
            session.invalidate();
        }
	}
    
    
    /**
     * Logout api.
     *
     * @param session the session
     */
    @RequestMapping({"/api/auth/logout"})
    @ResponseStatus(HttpStatus.NO_CONTENT)
    public void logoutApi(HttpSession session) {
        doLogout(session);
    }
    
    
    /**
     * Login form.
     *
     * @param locale the locale
     * @return the response entity
     */
    @RequestMapping("/auth/login")
    public ResponseEntity<String> loginForm(Locale locale, @RequestParam(name = "t", required = false) String t) {
    	if (StringUtils.hasText(t) && !t.startsWith("/")) {
    		t = null;
    	}
    	
    	HttpHeaders headers = new HttpHeaders();
    	headers.add("Content-Type", String.format("text/html; charset=%s", contentEncoding));
    	StringWriter sw = new StringWriter();	
    	templateService.evaluateTemplate("loginForm.html", getClass(), authHtmlParams(locale, messageSource, styleHref, t), locale, sw);
    	return new ResponseEntity<>(sw.toString(), headers, HttpStatus.OK);
    }

    
    /**
     * Signup form.
     *
     * @param locale the locale
     * @return the response entity
     */
    @RequestMapping("/auth/signup")
    public ResponseEntity<String> signupForm(Locale locale) {
    	if (!signupEnabled) {
    		throw new NotFoundException();
    	}
    	HttpHeaders headers = new HttpHeaders();
        headers.add("Content-Type", String.format("text/html; charset=%s", contentEncoding));
    	StringWriter sw = new StringWriter();
    	templateService.evaluateTemplate("signupForm.html", getClass(), authHtmlParams(locale, messageSource, styleHref, null), locale, sw);
    	return new ResponseEntity<>(sw.toString(), headers, HttpStatus.OK);
    }
    
    
    /**
     * Signup.
     *
     * @param signup the signup
     * @param locale the locale
     * @return the response entity
     */
    @RequestMapping(path = "/api/auth/signup", method = RequestMethod.POST)
    public ResponseEntity<List<AuthError>> signup(@RequestBody Signup signup, Locale locale) {
    	if (!signupEnabled) {
    		throw new NotFoundException();
    	}
    	//validate
    	if (signup == null) {
    		return new ResponseEntity<>(Collections.singletonList(new AuthError(null, "BadRequest")), HttpStatus.BAD_REQUEST);
    	}
    	
    	signup.setLocale(locale);
    	
    	List<AuthError> errors = new ArrayList<>();
    	
    	//TODO validate email
    	if (!StringUtils.hasText(signup.getUsername())) {
    		errors.add(new AuthError("username", "BadUsername"));
    	}
    	
    	if (!PasswordSecurity.validatePassword(signup.getPassword())) {
    		errors.add(new AuthError("password", "BadPassword"));
    	}
    	
    	try {
    		List<AuthError> errs = authService.register(signup);
    		if (errs == null || errs.isEmpty()) {
    			Authentication auth = authService.authenticate(signup.getUsername(), signup.getPassword());
    			SecurityContextHolder.getContext().setAuthentication(auth);
    		} else {
    			errors.addAll(errs);
    		}
    	} catch (Exception e) {
    		e.printStackTrace();
    		errors.add(new AuthError(null, e.getMessage()));
    	}
    	
    	//everything was OK - log user in
    	if (errors.isEmpty()) {
    		//return no errors
    		return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    	}
    	
		return new ResponseEntity<>(errors, HttpStatus.BAD_REQUEST);
    }
    
    
    @RequestMapping(path = "/api/auth/resetRequest", method = RequestMethod.POST)
    public ResponseEntity<List<AuthError>> resetToken(@RequestBody ResetPassword request) {
    	if (!resetEnabled) {
    		throw new NotFoundException();
    	}
    	authService.generateResetToken(request.getUsername());
    	return ResponseEntity.accepted().build();
    }
    
    
    @RequestMapping(path = "/api/auth/resetPassword", method = RequestMethod.POST)
    public ResponseEntity<List<AuthError>> reset(@RequestBody ResetPassword resetPassword) {
    	if (!resetEnabled) {
    		throw new NotFoundException();
    	}
    	String token = resetPassword.getToken();
    	if (!StringUtils.hasText(token)) {
    		return new ResponseEntity<>(Collections.singletonList(new AuthError(null, "BadRequest")), HttpStatus.BAD_REQUEST);
    	}
    	
    	List<AuthError> errors = new ArrayList<>();
    	
    	try {
    		List<AuthError> errs = authService.resetPassword(resetPassword);
    		if (errs == null || errs.isEmpty()) {
    			Authentication auth = authService.authenticate(resetPassword.getUsername(), resetPassword.getPassword());
    			SecurityContextHolder.getContext().setAuthentication(auth);
    		} else {
    			errors.addAll(errs);
    		}
    	} catch (Exception e) {
    		e.printStackTrace();
    		errors.add(new AuthError(null, e.getMessage()));
    	}
    	
    	if (errors.isEmpty()) {
    		//return no errors
    		return new ResponseEntity<>(HttpStatus.NO_CONTENT);
    	}
    	
		return new ResponseEntity<>(errors, HttpStatus.BAD_REQUEST);
    }
    
    
    @RequestMapping("/auth/reset")
    public ResponseEntity<String> resetForm(Locale locale) {
    	if (!resetEnabled) {
    		throw new NotFoundException();
    	}
    	HttpHeaders headers = new HttpHeaders();
        headers.add("Content-Type", String.format("text/html; charset=%s", contentEncoding));
    	StringWriter sw = new StringWriter();
    	templateService.evaluateTemplate("resetForm.html", getClass(), authHtmlParams(locale, messageSource, styleHref, null), locale, sw);
    	return new ResponseEntity<>(sw.toString(), headers, HttpStatus.OK);
    }
    
    
    /** The Constant ERRORS. */
    //
    static final String[] ERRORS = { 
        "UsernameRequired", 
        "PasswordRequired", 
        "UnexpectedError", 
        "UsernameTaken", 
        "BadUsername", 
        "BadPassword",
        "BadUsernameOrPassword",
        "ResetTokenSent"
    };
    
    /** The Constant ERROR_MSGS. */
    static final String[] ERROR_MSGS = Arrays.stream(ERRORS, 0, ERRORS.length).map(s -> { char c[] = s.toCharArray(); c[0] += 32; return String.format("auth.error.%s", new String(c)); }).collect(Collectors.toList()).toArray(new String[0]);
    

    /**
     * Auth html params.
     *
     * @param locale the locale
     * @param messageSource the message source
     * @param styleHref the style href
     * @return the map
     */
    public static Map<String, Object> authHtmlParams(Locale locale, MessageSource messageSource, String styleHref, String originalTarget) {
        Map<String, Object> params = new HashMap<>();
        params.put("_styleHref", styleHref);
        if (StringUtils.hasText(originalTarget)) {
        	params.put("loginTargetUri", originalTarget);
        } else {
        	params.put("loginTargetUri", "/");
        }
        StringBuilder sb = new StringBuilder("\t\t\twindow.__labels = [];\n");
        
        for (int i = 0; i < ERROR_MSGS.length; i++) {
            String s = ERROR_MSGS[i];
            String msg = messageSource.getMessage(s, null, s, locale);
            if (msg != null) {
                msg = msg.replaceAll("\\n", "\\\\n").replaceAll("\"", "\\\"");
                sb.append("\t\t\twindow.__labels['").append(ERRORS[i]).append("'] = \"").append(msg.trim()).append("\";\n");
            }
        }
        params.put("_labels", sb);
        
        //new params
        params.put("signupEnabled", signupEnabledStatic);
        params.put("resetEnabled", resetEnabledStatic);
        params.put("homeEnabled", homeEnabledStatic);
        return params;
    }
    
    
    /**
     * Sets the message source.
     *
     * @param messageSource the new message source
     */
    @Autowired(required = false)
    @Qualifier("i18nMessageSource")
    public void setMessageSource(MessageSource messageSource) {
        this.messageSource = messageSource;
    }


	@Override
	public void onApplicationEvent(ContextRefreshedEvent event) {
		signupEnabledStatic = signupEnabled;
		resetEnabledStatic = resetEnabled;
		homeEnabledStatic = homeEnabled;
	}
    
}
