/*
 * Created on 28. 7. 2017 15:26:38
 *
 */
package com.jh.common.web;

import java.io.IOException;
import java.io.InputStream;
import java.util.Objects;
import java.util.Scanner;
import java.util.function.Supplier;

import javax.annotation.PostConstruct;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ResourceLoaderAware;
import org.springframework.core.io.Resource;
import org.springframework.core.io.ResourceLoader;

/**
 * The Class ClientContent.
 *
 * @author Jan Hrabal
 */
public class ClientContent implements ResourceLoaderAware {

    /** The Constant LOG. */
    private static final Logger LOG = LoggerFactory.getLogger(ClientContent.class);
    
    /** The response modifiers. */
    private ClientContentModifier[] responseModifiers = new ClientContentModifier[0];

    /** The use cache. */
    private boolean useCache;
    
    /** The cached content. */
    private String cachedContent;
    
    /** The path. */
    private String path;
    
    /** The resource loader. */
    private ResourceLoader resourceLoader;
    
    /** The input stream supplier. */
    private Supplier<InputStream> inputStreamSupplier;

    /**
     * Instantiates a new client content.
     */
    public ClientContent() {
        super();
    }


    /**
     * Instantiates a new client content.
     *
     * @param resourceLoader the resource loader
     * @param path the path
     * @param useCache the use cache
     * @param responseModifiers the response modifiers
     */
    public ClientContent(ResourceLoader resourceLoader, String path, boolean useCache, ClientContentModifier...responseModifiers) {
        super();
        this.resourceLoader = resourceLoader;
        this.path = path;
        this.useCache = useCache;
        this.responseModifiers = responseModifiers;
        
        inputStreamSupplier = this::fetchResource;
    }

    
    /**
     * Instantiates a new client content.
     *
     * @param inputStreamSupplier the input stream supplier
     * @param useCache the use cache
     * @param responseModifiers the response modifiers
     */
    public ClientContent(Supplier<InputStream> inputStreamSupplier, boolean useCache, ClientContentModifier...responseModifiers) {
        super();
        this.useCache = useCache;
        this.responseModifiers = responseModifiers;
        this.inputStreamSupplier = inputStreamSupplier;
        
        if (responseModifiers != null) {
        	this.setResponseModifiers(responseModifiers);
        }
    }

    
    /**
     * Inits the.
     */
    @PostConstruct
    public void init() {
        inputStreamSupplier = this::fetchResource;
    }
    
    
    /**
     * Gets the content.
     *
     * @param request the request
     * @return the content
     */
    public String getContent(HttpServletRequest request) {
        String content = useCache && cachedContent != null ? cachedContent : (cachedContent = readContent());
        return evaluateIndexContent(content, request);
    }
    
    
    /**
     * Reads HTML content from provided path (classpath).
     *
     * @return the string
     */
    protected String readContent() {
        Objects.requireNonNull(inputStreamSupplier, "Input stream loader is not initialized");

        try (InputStream is = inputStreamSupplier.get()) {
            @SuppressWarnings("resource")
            Scanner s = new Scanner(is).useDelimiter("\\A");
            String content = s.hasNext() ? s.next() : null;
            s.close();
            
            return content;

        } catch (Exception e) {
            throw new IllegalStateException(String.format("Cannot read index file '%s'.", path), e);
        }
    }
    
    
    /**
     * Calls configured {@link ClientContentModifier}s.
     *
     * @param content the content
     * @param request the request
     * @return the string
     */
    protected String evaluateIndexContent(String content, HttpServletRequest request) {
        for (ClientContentModifier cirm : responseModifiers) {
            content = cirm.modifyResponse(content, request);
        }
        return content;
    }


    /**
     * Sets the response modifiers.
     *
     * @param responseModifiers the new response modifiers
     */
    public void setResponseModifiers(ClientContentModifier[] responseModifiers) {
        this.responseModifiers = responseModifiers;
    }


    /**
     * Sets the use cache.
     *
     * @param useCache the new use cache
     */
    public void setUseCache(boolean useCache) {
        this.useCache = useCache;
    }


    /**
     * Sets the path.
     *
     * @param path the new path
     */
    public void setPath(String path) {
        this.path = path;
    }


    /**
     * Sets the resource loader.
     *
     * @param arg0 the new resource loader
     * @see org.springframework.context.ResourceLoaderAware#setResourceLoader(org.springframework.core.io.ResourceLoader)
     */
    @Override
    @Autowired(required = false)
    public void setResourceLoader(ResourceLoader arg0) {
        resourceLoader = arg0;
    }
    
    
    /**
     * Fetch resource.
     *
     * @return the input stream
     */
    protected InputStream fetchResource() {
        if (resourceLoader == null) {
            LOG.warn("There is no resource loader defined. TODO");
            return null;
        }
        
        Resource resource = resourceLoader.getResource(path);
        if (!resource.exists()) {
            LOG.error(String.format("Cannot find resource '%s'", path));
            return null;
        }
        
        try {
            return resource.getInputStream();
        } catch (IOException e) {
            throw new RuntimeException(String.format("Cannot find resource '%s'", path), e);
        }
    };
    
    
    
    /**
     * Render.
     *
     * @param content the content
     * @param response the response
     */
    public static void render(String content, HttpServletResponse response) {
        response.setContentType("text/html");
        response.setCharacterEncoding("UTF-8");
        try {
            response.getWriter().print(content);
        } catch (IOException e) {
            throw new RuntimeException("Cannot render response", e);
        }
        
        response.setStatus(HttpServletResponse.SC_OK);
    }
    
}
